﻿/*****************************************************************************
*                                                                            *
*  XELF.OpenNI (OpenNI .NET Wrapper / Specialized)                           *
*  Copyright © 2010-2011 XELF                                                *
*  http://xelf.info/                                                         *
*                                                                            *
*  This file is part of XELF.OpenNI                                          *
*                                                                            *
*  XELF.OpenNI is free software: you can redistribute it and/or modify       *
*  it under the terms of the GNU Lesser General Public License as published  *
*  by the Free Software Foundation, either version 3 of the License, or      *
*  (at your option) any later version.                                       *
*                                                                            *
*  XELF.OpenNI is distributed in the hope that it will be useful,            *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              *
*  GNU Lesser General Public License for more details.                       *
*                                                                            *
*  You should have received a copy of the GNU Lesser General Public License  *
*  along with XELF.OpenNI. If not, see <http://www.gnu.org/licenses/>.       *
*                                                                            *
*****************************************************************************/

#pragma once

namespace XELF { namespace OpenNI {
	using namespace System;

	public enum struct Status {
		Ok = XN_STATUS_OK,
	};

	public enum struct SkeletonJoint {
		None = 0,
		Head = XN_SKEL_HEAD,
		Neck = XN_SKEL_NECK,
		Torso = XN_SKEL_TORSO,
		Waist = XN_SKEL_WAIST,
		LeftCollar = XN_SKEL_LEFT_COLLAR,
		LeftShoulder = XN_SKEL_LEFT_SHOULDER,
		LeftElbow = XN_SKEL_LEFT_ELBOW,
		LeftWrist = XN_SKEL_LEFT_WRIST,
		LeftHand = XN_SKEL_LEFT_HAND,
		LeftFingertip = XN_SKEL_LEFT_FINGERTIP,
		RightCollar = XN_SKEL_RIGHT_COLLAR,
		RightShoulder = XN_SKEL_RIGHT_SHOULDER,
		RightElbow = XN_SKEL_RIGHT_ELBOW,
		RightWrist = XN_SKEL_RIGHT_WRIST,
		RightHand = XN_SKEL_RIGHT_HAND,
		RightFingertip = XN_SKEL_RIGHT_FINGERTIP,
		LeftHip = XN_SKEL_LEFT_HIP,
		LeftKnee = XN_SKEL_LEFT_KNEE,
		LeftAnkle = XN_SKEL_LEFT_ANKLE,
		LeftFoot = XN_SKEL_LEFT_FOOT,
		RightHip = XN_SKEL_RIGHT_HIP,
		RightKnee = XN_SKEL_RIGHT_KNEE,
		RightAnkle = XN_SKEL_RIGHT_ANKLE,
		RightFoot = XN_SKEL_RIGHT_FOOT,
	};

	public enum struct Capability {
		AlternativeViewPoint = 1 << 0,
		Cropping = 1 << 1,
		ErrorState = 1 << 2,
		ExtendedSerialization = 1 << 3,
		FrameSync = 1 << 4,
		LockAware = 1 << 5,
		Mirror = 1 << 6,
		UserPoseDetection = 1 << 7,
		UserSkeleton = 1 << 8,
		UserPosition = 1 << 9,
	};

	// XnVector3D / XnPoint3D
	public value struct Vector3 {
	public:
		Vector3(float x, float y, float z) { X = x; Y = y; Z = z; }
	public:
		float X;
	public:
		float Y;
	public:
		float Z;

	public:
		virtual String^ ToString() override {
			return String::Format("{0}, {1}, {2}", X, Y, Z);
		}
	};

	// XnMatrix3X3
	public value struct Matrix3x3 {
	public:
		float M11, M12, M13;
		float M21, M22, M23;
		float M31, M32, M33;
	};

	// XnPlane3D
	public value struct PositionNormal {
	public:
		Vector3 Normal;
	public:
		Vector3 Position;
	public:
		virtual String^ ToString() override {
			return String::Format("Normal = {{{0}}}, Position = {{{1}}},", Normal, Position);
		}
	};

	public value struct SkeletonJointPosition {
	public:
		Vector3 Position;
	public:
		float Confidence;
	};

	public value struct SkeletonJointOrientation {
	public:
		Matrix3x3 Orientation;
	public:
		float Confidence;
	};

	public value struct SkeletonJointTransform {
	public:
		SkeletonJointPosition Position;
	public:
		SkeletonJointOrientation Orientation;
	public:
		float Confidence;
	};

	public value struct MapOutputMode {
	public:
		UInt32 Width;
	public:
		UInt32 Height;
	public:
		UInt32 FramesPerSecond;
	public:
		virtual String^ ToString() override {
			return String::Format("{0}, {1}, {2}", Width, Height, FramesPerSecond);
		}
	};

	public value struct FieldOfView {
	public:
		// [radians]
		double Horizontal;
	public:
		// [radians]
		double Vertical;

	public:
		virtual String^ ToString() override {
			return Horizontal.ToString() + ", " + Vertical.ToString();
		}
	};

	public value struct DepthMetaData {
	public:
		IntPtr Map;
	public:
		IntPtr Data;
	public:
		UInt16 Depth; // ZResolution;
	};

	public value struct SceneMetaData {
	public:
		IntPtr Map; // XnMapMetaData
	public:
		IntPtr Labels; // XnLabel
	};

} }
