﻿/*****************************************************************************
*                                                                            *
*  XELF.OpenNI (OpenNI .NET Wrapper / Specialized)                           *
*  Copyright © 2010-2011 XELF                                                *
*  http://xelf.info/                                                         *
*                                                                            *
*  This file is part of XELF.OpenNI                                          *
*                                                                            *
*  XELF.OpenNI is free software: you can redistribute it and/or modify       *
*  it under the terms of the GNU Lesser General Public License as published  *
*  by the Free Software Foundation, either version 3 of the License, or      *
*  (at your option) any later version.                                       *
*                                                                            *
*  XELF.OpenNI is distributed in the hope that it will be useful,            *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              *
*  GNU Lesser General Public License for more details.                       *
*                                                                            *
*  You should have received a copy of the GNU Lesser General Public License  *
*  along with XELF.OpenNI. If not, see <http://www.gnu.org/licenses/>.       *
*                                                                            *
*****************************************************************************/

#include <XnContext.h>
#include <XnPrdNode.h>
#include <XnLicensing.h>
#include <XnEnumerationErrors.h>
#include <XnNodeAllocator.h>
#include "Types.h"

using namespace System;
using namespace System::Collections;
using namespace System::Collections::Generic;
using namespace System::Runtime::InteropServices;

namespace XELF { namespace OpenNI {

value class ErrorCollectionIterator;

public ref class ErrorCollection : public IDisposable {
internal:
	XnEnumerationErrors* _;

	static void Check(XnStatus status) {
		if (status != XN_STATUS_OK) {
			throw gcnew Exception(String::Format("{0,0:X8}: {1}",
				(int)status, gcnew String(xnGetStatusString(status))));
		}
	}
public:
	ErrorCollection() {
		XnEnumerationErrors* _;
		XnStatus status = xnEnumerationErrorsAllocate(&_);
		Check(status);
		this->_ = _;
	}
	ErrorCollectionIterator GetEnumerator();
	~ErrorCollection() {
		xnEnumerationErrorsFree(_);
		_ = nullptr;
	}
	void Clear() {
		XnStatus status = xnEnumerationErrorsClear(_);
		Check(status);
	}
};

public value class ErrorCollectionIterator : public System::Collections::Generic::IEnumerator<Status> {
internal:
	ErrorCollection^ _;
	XnEnumerationErrorsIterator iterator;
public:
	ErrorCollectionIterator(ErrorCollection^ _) {
		this->_ = _;
		Reset();
	}
	virtual void Reset() {
		iterator = xnEnumerationErrorsGetFirst(_->_);
	}
	virtual property Status Current {
		Status get() {
			if (iterator == nullptr) return Status::Ok;
			return (Status)xnEnumerationErrorsGetCurrentError(iterator);
		}
	}
	virtual property Object^ CurrentBase {
		virtual Object^ get() = System::Collections::IEnumerator::Current::get {
			if (iterator == nullptr) return Status::Ok;
			return (Status)xnEnumerationErrorsGetCurrentError(iterator);
		}
	}
	virtual bool MoveNext() {
		if (iterator == nullptr) return false;
		iterator = xnEnumerationErrorsGetNext(iterator);
		return true;
	}
};

ErrorCollectionIterator ErrorCollection::GetEnumerator() {
	return ErrorCollectionIterator(this);
}

#include <string.h>

ref class StringToCharPtr : IDisposable {
	IntPtr _;
public:
	StringToCharPtr(String^ text) {
		_ = Marshal::StringToHGlobalAnsi(text);
	}
	~StringToCharPtr() {
		Marshal::FreeHGlobal(_);
	}
	char* ToPointer() { return (char*)(void*)_; }
};

public ref class Context {
internal:
	XnContext* _;
public:
	Context(String^ configXmlFilePath, ErrorCollection^ errors) {
		StringToCharPtr text(configXmlFilePath);
		XnContext* context;
		XnStatus result = xnInitFromXmlFile(text.ToPointer(), &context, errors ? errors->_ : nullptr);
		Context::Check((Status)result);
		_ = context;
	}
	Context() {
		XnContext* context;
		XnStatus result = xnInit(&context);
		Context::Check((Status)result);
		_ = context;
	}

public:
	Status Start() {
		return (Status)xnStartGeneratingAll(_);
	}

public:
	Status Stop() {
		return (Status)xnStopGeneratingAll(_);
	}

public:
	void WaitAndUpdateAll() {
		XnStatus status = xnWaitAndUpdateAll(_);
		ErrorCollection::Check(status);
	}

public:
	virtual ~Context() {
		xnShutdown(_);
		_ = nullptr;
	}

public:
	static void Check(Status status) {
		if (status != Status::Ok) {
			throw gcnew Exception(String::Format("{0,0:X8}: {1}",
				(int)status, gcnew String(xnGetStatusString((XnStatus)status))));
		}
	}
private:
	static void Check(XnStatus status) {
		if (status != XN_STATUS_OK) {
			throw gcnew Exception(String::Format("{0,0:X8}: {1}",
				(int)status, gcnew String(xnGetStatusString(status))));
		}
	}
};

public ref class Node abstract {
internal:
	XnNodeHandle _;
	Context^ context;

public:
	bool IsCapabilitySupported(Capability oneOfCapability) {
		XnBool result = false;
		switch(oneOfCapability) {
		case Capability::AlternativeViewPoint:
			result = xnIsCapabilitySupported(_, XN_CAPABILITY_ALTERNATIVE_VIEW_POINT);
		case Capability::Cropping:
			result = xnIsCapabilitySupported(_, XN_CAPABILITY_CROPPING);
		case Capability::ErrorState:
			result = xnIsCapabilitySupported(_, XN_CAPABILITY_ERROR_STATE);
		case Capability::ExtendedSerialization:
			result = xnIsCapabilitySupported(_, XN_CAPABILITY_EXTENDED_SERIALIZATION);
		case Capability::FrameSync:
			result = xnIsCapabilitySupported(_, XN_CAPABILITY_FRAME_SYNC);
		case Capability::LockAware:
			result = xnIsCapabilitySupported(_, XN_CAPABILITY_LOCK_AWARE);
		case Capability::Mirror:
			result = xnIsCapabilitySupported(_, XN_CAPABILITY_MIRROR);
		case Capability::UserPoseDetection:
			result = xnIsCapabilitySupported(_, XN_CAPABILITY_POSE_DETECTION);
		case Capability::UserSkeleton:
			result = xnIsCapabilitySupported(_, XN_CAPABILITY_SKELETON);
		case Capability::UserPosition:
			result = xnIsCapabilitySupported(_, XN_CAPABILITY_USER_POSITION);
		}
		return result ? true : false;
	}
};

public ref class Generator abstract : Node {
public:
	Status Wait() {
		return (Status)xnWaitOneUpdateAll(context->_, _);
	}
	property XELF::OpenNI::MapOutputMode MapOutputMode {
		XELF::OpenNI::MapOutputMode get() {
			XELF::OpenNI::MapOutputMode result;
			XnStatus status = xnGetMapOutputMode(_, (XnMapOutputMode*)&result);
			ErrorCollection::Check(status);
			return result;
		}
		void set(XELF::OpenNI::MapOutputMode mode) {
			XnStatus status = xnSetMapOutputMode(_, (XnMapOutputMode*)&mode);
			ErrorCollection::Check(status);
		}
	}

#pragma region Alternative View Point Capability

public:
	property Generator^ ViewPoint {
		void set(Generator^ other) {
			if (other == nullptr) {
				XnStatus status = xnResetViewPoint(_);
				ErrorCollection::Check(status);
			} else {
				XnStatus status = xnSetViewPoint(_, other->_);
				ErrorCollection::Check(status);
			}
		}
	}

public:
	bool IsViewPointAs(Generator^ other) {
		return xnIsViewPointAs(_, other->_) ? true : false;
	}

public:
	bool IsViewPointSupported(Generator^ other) {
		return xnIsViewPointSupported(_, other->_) ? true : false;
	}

#pragma endregion

};

public ref class SceneAnalyzer : Node {
public:
	SceneAnalyzer(Context^ context, ErrorCollection^ errors) {
		XnNodeHandle node;
		this->context = context;
		XnStatus status = xnCreateSceneAnalyzer(context->_, &node, nullptr, errors ? errors->_ : nullptr);
		Context::Check((Status)status);
		_ = node;
	}
public:
	property Nullable<PositionNormal> Floor {
		Nullable<PositionNormal> get() {
			PositionNormal result;
			XnStatus status = xnGetFloor(_, (XnPlane3D*)&result);
			if (status == XN_STATUS_OK) 
				return result;
			return Nullable<PositionNormal>();
		}
	}
	property SceneMetaData MetaData {
		SceneMetaData get() {
			SceneMetaData data;
			xnGetSceneMetaData(_, (XnSceneMetaData*)&data);
			return data;
		}
	}
	property IntPtr LabelMap {
		IntPtr get() {
			return IntPtr((void*)xnGetLabelMap(_));
		}
	}
};

public ref class ImageGenerator : Generator {
public:
	static ImageGenerator^ Find(Context^ context) {
		XnNodeHandle generator;
		XnStatus status = xnFindExistingNodeByType(context->_, XN_NODE_TYPE_IMAGE, &generator);
		Context::Check((Status)status);
		return gcnew ImageGenerator(context, generator);
	}
private:
	ImageGenerator(Context^ context, XnNodeHandle imageGenerator) {
		this->context = context;
		_ = imageGenerator;
	}
public:
	ImageGenerator(Context^ context) {
		XnNodeHandle generator;
		XnStatus status = xnFindExistingNodeByType(context->_, XN_NODE_TYPE_IMAGE, &generator);
		Context::Check((Status)status);
		this->context = context;
		_ = generator;
	}
public:
	property IntPtr Map {
		IntPtr get() {
			return IntPtr(xnGetImageMap(_));
		}
	}
};

static void XN_CALLBACK_TYPE UserEnteredCallback(XnNodeHandle handle, XnUserID user, void* cookie);
static void XN_CALLBACK_TYPE UserLeftCallback(XnNodeHandle handle, XnUserID user, void* cookie);
static void XN_CALLBACK_TYPE UserCalibrationStartCallback(XnNodeHandle handle, XnUserID user, void* cookie);
static void XN_CALLBACK_TYPE UserCalibrationEndCallback(XnNodeHandle handle, XnUserID user, XnBool success, void* cookie);

public ref class UserGenerator : Generator {
private:
	GCHandle holder;
public:
	static UserGenerator^ Find(Context^ context) {
		XnNodeHandle generator;
		XnStatus status = xnFindExistingNodeByType(context->_, XN_NODE_TYPE_USER, &generator);
		ErrorCollection::Check(status);
		return gcnew UserGenerator(context, generator);
	}
private:
	UserGenerator(Context^ context, XnNodeHandle userGenerator) {
		this->context = context;
		_ = userGenerator;
		RegisterUserCallback();
	}
public:
	UserGenerator(Context^ context, ErrorCollection^ errors) {
		XnNodeHandle generator;
		XnStatus status = xnCreateUserGenerator(context->_, &generator, nullptr, errors ? errors->_ : nullptr);
		ErrorCollection::Check(status);
		this->context = context;
		_ = generator;
		RegisterUserCallback();
	}
	virtual ~UserGenerator() {
		if (_) {
			UnregisterUserCallback();
			_ = nullptr;
		}
	}

	property UInt16 Count {
		UInt16 get() {
			return xnGetNumberOfUsers(_);
		}
	}
	int GetUsers(cli::array<UInt16>^ users) {
		pin_ptr<UInt16> p = &users[0];
		UInt16 numberUsers = users->Length;
		XnStatus status = xnGetUsers(_, (XnUserID*)p, &numberUsers);
		ErrorCollection::Check(status);
		return numberUsers;
	}
public:
	bool IsSkeletonTracking(UInt16 user) {
		return xnIsSkeletonTracking(_, user) != 0;
	}
	void StartSkeletonTracking(UInt16 user) {
		XnStatus status = xnStartSkeletonTracking(_, user);
		ErrorCollection::Check(status);
	}
	void StopSkeletonTracking(UInt16 user) {
		XnStatus status = xnStopSkeletonTracking(_, user);
		ErrorCollection::Check(status);
	}
	void ResetSkeleton(UInt16 user) {
		XnStatus status = xnResetSkeleton(_, user);
		ErrorCollection::Check(status);
	}
	void RequestSkeletonCalibration(UInt16 user, bool force) {
		XnStatus status = xnRequestSkeletonCalibration(_, user, force);
		ErrorCollection::Check(status);
	}
	void SetSkeletonSmoothing(float factor) {
		XnStatus status = xnSetSkeletonSmoothing(_, factor);
		ErrorCollection::Check(status);
	}
	Nullable<Vector3> GetUserCenterOfMass(UInt16 user) {
		Vector3 result;
		XnStatus status = xnGetUserCoM(_, user, (XnPoint3D*)&result);
		if (status == XN_STATUS_OK)
			return result;
		return Nullable<Vector3>();
	}
	SceneMetaData GetUserPixels(UInt16 user) {
		SceneMetaData data;
		xnGetUserPixels(_, user, (XnSceneMetaData*)&data);
		return data;
	}

private:
	XnCallbackHandle userCallbackHandle;
	XnCallbackHandle calibrationCallbackHandle;

	void RegisterUserCallback() {
		holder = GCHandle::Alloc(this);
		XnCallbackHandle userCallback;
		XnCallbackHandle calibrationCallback;
		xnRegisterUserCallbacks(_, UserEnteredCallback, UserLeftCallback, (void*)(IntPtr)holder, &userCallback);
		xnRegisterCalibrationCallbacks(_, UserCalibrationStartCallback, UserCalibrationEndCallback, (void*)(IntPtr)holder, &calibrationCallback);
		userCallbackHandle = userCallback;
		calibrationCallbackHandle = calibrationCallback;
	}
	void UnregisterUserCallback() {
		xnUnregisterUserCallbacks(_, userCallbackHandle);
		xnUnregisterCalibrationCallbacks(_, calibrationCallbackHandle);
		holder.Free();
	}

public:
	event Action<UInt16>^ UserEntered;
	event Action<UInt16>^ UserLeft;
	event Action<UInt16>^ CalibrationStart;
	event Action<UInt16, bool>^ CalibrationEnd;

internal:
	void PerformUserEntered(XnUserID user) {
		UserEntered(user);
	}
	void PerformUserLeft(XnUserID user) {
		UserLeft(user);
	}
	void PerformCalibrationStart(XnUserID user) {
		CalibrationStart(user);
	}
	void PerformCalibrationEnd(XnUserID user, XnBool success) {
		CalibrationEnd(user, success != 0 ? true : false);
	}

public:
	UInt16 GetActiveJoints(cli::array<SkeletonJoint>^ joints) {
		pin_ptr<SkeletonJoint> p = &joints[0];
		UInt16 count = (UInt16)min(0xffff, joints->Length);
		XnStatus status = xnEnumerateActiveJoints(_, (XnSkeletonJoint*)p, &count);
		ErrorCollection::Check(status);
		return count;
	}
	void SetJointActive(SkeletonJoint joint, bool active) {
		XnStatus status = xnSetJointActive(_, (XnSkeletonJoint)joint, active);
		ErrorCollection::Check(status);
	}

#pragma region Skeleton Capability
public:

	Nullable<SkeletonJointTransform> GetSkeletonJoint(UInt16 user, SkeletonJoint joint) {
		SkeletonJointTransform result;
		XnStatus status = xnGetSkeletonJoint(_, user, (XnSkeletonJoint)joint, (XnSkeletonJointTransformation*)&result);
		if (status == XN_STATUS_OK)
			return result;
		return Nullable<SkeletonJointTransform>();
	}

	Nullable<SkeletonJointOrientation> GetSkeletonJointOrientation(UInt16 user, SkeletonJoint joint) {
		SkeletonJointOrientation result;
		XnStatus status = xnGetSkeletonJointOrientation(_, user, (XnSkeletonJoint)joint, (XnSkeletonJointOrientation*)&result);
		if (status == XN_STATUS_OK)
			return result;
		return Nullable<SkeletonJointOrientation>();
	}

#pragma endregion

};

static void XN_CALLBACK_TYPE UserEnteredCallback(XnNodeHandle handle, XnUserID user, void* cookie) {
	GCHandle holder = GCHandle::FromIntPtr(IntPtr(cookie));
	safe_cast<UserGenerator^>(holder.Target)->PerformUserEntered(user);
}
static void XN_CALLBACK_TYPE UserLeftCallback(XnNodeHandle handle, XnUserID user, void* cookie) {
	GCHandle holder = GCHandle::FromIntPtr(IntPtr(cookie));
	safe_cast<UserGenerator^>(holder.Target)->PerformUserLeft(user);
}
static void XN_CALLBACK_TYPE UserCalibrationStartCallback(XnNodeHandle handle, XnUserID user, void* cookie) {
	GCHandle holder = GCHandle::FromIntPtr(IntPtr(cookie));
	safe_cast<UserGenerator^>(holder.Target)->PerformCalibrationStart(user);
}
static void XN_CALLBACK_TYPE UserCalibrationEndCallback(XnNodeHandle handle, XnUserID user, XnBool success, void* cookie) {
	GCHandle holder = GCHandle::FromIntPtr(IntPtr(cookie));
	safe_cast<UserGenerator^>(holder.Target)->PerformCalibrationEnd(user, success);
}

public ref class DepthGenerator : Generator {
public:
	static DepthGenerator^ Find(Context^ context) {
		XnNodeHandle depthGenerator;
		XnStatus status = xnFindExistingNodeByType(context->_, XN_NODE_TYPE_DEPTH, &depthGenerator);
		Context::Check((Status)status);
		return gcnew DepthGenerator(context, depthGenerator);
	}

private:
	DepthGenerator(Context^ context, XnNodeHandle depthGenerator) {
		this->context = context;
		_ = depthGenerator;
	}

public:
	DepthGenerator(Context^ context) {
		XnNodeHandle depthGenerator;
		XnNodeQuery* query = nullptr;
		XnEnumerationErrors* errors = nullptr;
		XnStatus result = xnCreateDepthGenerator(context->_, &depthGenerator, query, errors);
		Context::Check((Status)result);
		this->context = context;
		_ = depthGenerator;
	}

public:
	property IntPtr Map {
		// XnDepthPixel*
		IntPtr get() {
			return IntPtr(xnGetDepthMap(_));
		}
	}

public:
	property DepthMetaData MetaData {
		DepthMetaData get() {
			DepthMetaData result;
			xnGetDepthMetaData(_, (XnDepthMetaData*)&result);
			return result;
		}
	}

public:
	property FieldOfView FieldOfView {
		XELF::OpenNI::FieldOfView get() {
			XELF::OpenNI::FieldOfView fov;
			XnStatus result = xnGetDepthFieldOfView(_, (XnFieldOfView*)&fov);
			Context::Check((Status)result);
			return fov;
		}
	}

public:
	void ConvertProjectiveToRealWorld(cli::array<Vector3>^ projectivePoints, int count, cli::array<Vector3>^ realWorldPoints) {
		if (projectivePoints->Length < count) throw gcnew System::ArgumentException();
		if (realWorldPoints->Length < count) throw gcnew System::ArgumentException();
		pin_ptr<Vector3> s = &projectivePoints[0];
		pin_ptr<Vector3> d = &realWorldPoints[0];
		xnConvertProjectiveToRealWorld(_, count, (XnPoint3D*)s, (XnPoint3D*)d);
	}

public:
	void ConvertRealWorldToProjective(cli::array<Vector3>^ realWorldPoints, int count, cli::array<Vector3>^ projectivePoints) {
		if (projectivePoints->Length < count) throw gcnew System::ArgumentException();
		if (realWorldPoints->Length < count) throw gcnew System::ArgumentException();
		pin_ptr<Vector3> s = &realWorldPoints[0];
		pin_ptr<Vector3> d = &projectivePoints[0];
		xnConvertRealWorldToProjective(_, count, (XnPoint3D*)s, (XnPoint3D*)d);
	}
};

} } // namespace XELF::OpenNI
